package moraxus.server;

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import moraxus.server.storage.Content;
import moraxus.server.storage.MoraxusKey;
import moraxus.server.storage.Storage;
import moraxus.server.user.Authenticator;
import moraxus.server.user.User;

import com.dyuproject.openid.RelyingParty;

@SuppressWarnings("serial")
public class ContentServlet extends HttpServlet {

	public void doGet(HttpServletRequest req, HttpServletResponse resp) throws IOException {
		// Find the path, for instance /content/blog
		String path = Util.getFormattedPath(req.getPathInfo());
		log("GET request to " + path);
		// Find the entity with the given path
		Storage db = Configuration.getStorage();
		MoraxusKey key = db.getKeyFromPath(path);
		
		// No key found?
		if (key == null) {
			OutputWriter.write(resp, HttpServletResponse.SC_BAD_REQUEST, req.getParameter("format"), new OutMessage("error", "no key found"));
			return;
		}
		log("Key found for path " + path + ". Key is " + key);
		
		// No access?
		if(!Util.hasReadAccess(req, key)){
			OutputWriter.write(resp, HttpServletResponse.SC_UNAUTHORIZED, req.getParameter("format"), new OutMessage("error", "forbidden"));
		}	
		
		try {
			String type = req.getParameter("type");
			if (type != null) {// Perform query
				List<Content> result = db.search(key, type);

				// Should we load the children too?
				String loadChildren = req.getParameter("children");
				if (loadChildren != null)
					if (loadChildren.equals("true") || loadChildren.equals("1") || loadChildren.equals("yes"))
						for (Content c : result)
							c.setChildren(db.getChildrenKeys(c.getKey()));

				OutputWriter.write(resp, HttpServletResponse.SC_ACCEPTED, req.getParameter("format"), result);
			} else {// Print object data
				Content c = db.getContent(key);

				// Add children references
				c.setChildren(db.getChildrenKeys(key));

				OutputWriter.write(resp, HttpServletResponse.SC_ACCEPTED, req.getParameter("format"), c);
			}
		} catch (Exception e) {
			OutputWriter.write(resp, HttpServletResponse.SC_INTERNAL_SERVER_ERROR, req.getParameter("format"), new OutMessage("error", e.toString()));
			log("Error: " + e);
		}
	}

	@SuppressWarnings("rawtypes")
	public void doPost(HttpServletRequest req, HttpServletResponse resp) throws IOException {
		// First we check for a valid API key
		String apiKey = req.getHeader("apiKey");
		
		if (!Util.isValidAPIKey(apiKey)) {
			OutputWriter.write(resp, HttpServletResponse.SC_FORBIDDEN, req.getParameter("format"), new OutMessage("error", "invalid API key"));
			log("Error: invalid api key");
			return;
		}

		// Find the path, for instance /content/blog and saving it for later use
		String path = Util.getFormattedPath(req.getPathInfo());
		String originalPath = path;
		log("POST request to " + path);
		// No type, no action!
		String type = req.getParameter("type");
		if (type == null || type.equals("")) {
			OutputWriter.write(resp, HttpServletResponse.SC_BAD_REQUEST, req.getParameter("format"), new OutMessage("error", "no type provided"));
			return;
		}

		// Find the entity with the given path
		Storage db = Configuration.getStorage();
		MoraxusKey pathKey = db.getKeyFromPath(path);
		Stack<String> newEntities = new Stack<String>();

		// No key? Then we have to create the entity (possibly recursively back
		// to the root directory)
		while (pathKey == null) {
			newEntities.push(path.equals("") ? "/" : path);

			// Break if we are at the root
			if (path.equals(""))
				break;

			path = path.substring(0, path.lastIndexOf("/"));
			pathKey = db.getKeyFromPath(path);
		}

		// Create missing entities if needed
		while (newEntities.size() > 0) {
			path = newEntities.pop();

			// Creating node
			Content c = null;
			if (path.equals("/")) {
				c = new Content("collection", apiKey, "-1");
				c.setProperty("name", "content");
			} else {
				c = new Content("collection", apiKey, "-1", pathKey);
				c.setProperty("name", path.substring(path.lastIndexOf("/") + 1));
			}

			MoraxusKey prevParent = pathKey;
			pathKey = db.saveObject(c);

			// Save path/key pair
			c = new Content("path_key", apiKey, "-1");
			c.setProperty("path", path);
			c.setProperty("key", pathKey);
			db.saveObject(c);

			// Save the parentKey/key pair unless it is the root entity
			if (!path.equals("/")) {
				c = new Content("parentKey_key", apiKey, "-1");
				c.setProperty("parentKey", prevParent);
				c.setProperty("key", pathKey);
				db.saveObject(c);
			}
		}
		
		//Hmm, do you need access to post to a collection?
//		if(!Util.hasWriteAccess(req, pathKey)){
//			OutputWriter.write(resp, HttpServletResponse.SC_UNAUTHORIZED, req.getParameter("format"), new OutMessage("error", "forbidden"));
//		}
		
		// Get the user id
		Authenticator auth=Configuration.getAuthenticator();
		User user=auth.getCurrentUser(req);
		String userId;
		if(user==null)
			userId="-1";
		else
			userId=user.getId();

		// Create a new content object
		Content content = new Content(type, apiKey, userId, pathKey);

		// Add properties
		Map map = req.getParameterMap();
		Iterator it = map.entrySet().iterator();

		while (it.hasNext()) {
			Map.Entry pair = (Map.Entry) it.next();
			String key = pair.getKey().toString();
			String value = req.getParameter(key);
			content.setProperty(key, value);
		}

		// Save the entity and get the key
		MoraxusKey key = db.saveObject(content);

		// Save the path/key pair
		originalPath += "/" + key;
		Content c = new Content("path_key", apiKey, userId);
		c.setProperty("path", originalPath);
		c.setProperty("key", key);
		db.saveObject(c);

		// Save the parentKey/key pair
		c = new Content("parentKey_key", apiKey, userId);
		c.setProperty("parentKey", pathKey);
		c.setProperty("key", key);
		db.saveObject(c);
		
		// Give the user write access to the object
		if(user!=null){
			db.setWriteAccess(user.getId(), key.toString());
		}

		// Print the key
		OutputWriter.write(resp, HttpServletResponse.SC_ACCEPTED, req.getParameter("format"), new OutMessage("key", key.toString()));
	}
	
	public void doDelete(HttpServletRequest req, HttpServletResponse resp) throws IOException {
		// Find the path, for instance /content/blog
		String path = Util.getFormattedPath(req.getPathInfo());
		
		// Find the entity with the given path
		Storage db = Configuration.getStorage();
		MoraxusKey key = db.getKeyFromPath(path);
		
		// No key found?
		if (key == null) {
			OutputWriter.write(resp, HttpServletResponse.SC_BAD_REQUEST, req.getParameter("format"), new OutMessage("error", "no key found"));
			return;
		}
		log("Key found for path " + path + ". Key is " + key);
		
		LinkedList<MoraxusKey> toBeDeleted=new LinkedList<MoraxusKey>();
		
		toBeDeleted.add(key);
		
		// Delete object and recursively all children we have writeaccess to
		while(toBeDeleted.size()>0){
			key=toBeDeleted.remove();
			if(Util.hasWriteAccess(req, key)){
				log("Deleting key " + key);
				toBeDeleted.addAll(db.getChildrenKeys(key));
				db.delete(key);
			}
		}
		
		
		// No access?
		if(!Util.hasWriteAccess(req, key)){
			OutputWriter.write(resp, HttpServletResponse.SC_UNAUTHORIZED, req.getParameter("format"), new OutMessage("error", "forbidden"));
		}	
	}

	public void log(String msg) {
		System.out.println(msg);
	}
}
