package moraxus.server.storage;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.google.appengine.api.datastore.DatastoreService;
import com.google.appengine.api.datastore.DatastoreServiceFactory;
import com.google.appengine.api.datastore.Entity;
import com.google.appengine.api.datastore.EntityNotFoundException;
import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.KeyFactory;
import com.google.appengine.api.datastore.PreparedQuery;
import com.google.appengine.api.datastore.Query;
import com.google.appengine.api.datastore.Query.FilterOperator;
import com.google.appengine.api.datastore.Text;

public class AppEngineStorage implements Storage {
	private DatastoreService datastore = DatastoreServiceFactory.getDatastoreService();

	@Override
	public MoraxusKey getKeyFromPath(String path) {
		Query q = new Query("path_key");
		q.addFilter("path", Query.FilterOperator.EQUAL, path);

		PreparedQuery pq = datastore.prepare(q);

		for (Entity result : pq.asIterable()) {
			String keyString = (String) result.getProperty("key");
			return new MoraxusKey(keyString);
		}	
		
		// No entity found at the given path
		return null;
	}
	
	@Override
	public List<Content> search(MoraxusKey ancestor, String type) {
		return search(ancestor, type, new HashMap<String, String>());
	}

	@Override
	public List<Content> search(MoraxusKey ancestor, String type, Map<String, String> parameters) {
		List<Content> list=new ArrayList<Content>();
		Query q = new Query(type);
		
		if(ancestor!=null){
			Key ancestorKey=KeyFactory.stringToKey(ancestor.toString());
			q.setAncestor(ancestorKey);
		}
		//Add parameters
		Iterator it = parameters.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry pair = (Map.Entry) it.next();
			String key = pair.getKey().toString();
			String value = pair.getValue().toString();
			q.addFilter(key, FilterOperator.EQUAL, value);
		}
		
		PreparedQuery pq = datastore.prepare(q);

		for (Entity e : pq.asIterable()) {
			if(!e.getKey().equals(ancestor)){
				//Create a content object from the entity
				list.add(entityToContent(e));
			}
		}	
		
		return list;
	}

	@Override
	public Content getContent(MoraxusKey key) {
		Entity e=null;
		try{
			e=datastore.get(KeyFactory.stringToKey(key.toString()));
		}catch(EntityNotFoundException ex){
			
		}
		if(e==null)
			return null;
		
		return entityToContent(e);
	}
	
	private Content entityToContent(Entity e){
		Content c=new Content(e.getKind(), e.getProperty("apiKey").toString(), e.getProperty("creatorId").toString());
		
		Map<String, Object> map = e.getProperties();
		Iterator it = map.entrySet().iterator();

		//Add properties
		while (it.hasNext()) {
			Map.Entry pair = (Map.Entry) it.next();
			String key = pair.getKey().toString();
			if(pair.getValue() instanceof Text){
				Text t=(Text) pair.getValue();
				c.setProperty(key, t.getValue());
			}else{
				String value = pair.getValue().toString();
				c.setProperty(key, value);
			}
		}
		c.setKey(new MoraxusKey(KeyFactory.keyToString(e.getKey())));
		return c;
	}
	
	private Entity contentToEntity(Content c){
		Entity e=null;
		if(c.getProperty("parentKey")!=null){
			e=new Entity(c.getProperty("type"), KeyFactory.stringToKey(c.getProperty("parentKey")));
		}else{
			e=new Entity(c.getProperty("type"));
		}
		
		Map<String, String> map = c.getProperties();
		Iterator it = map.entrySet().iterator();

		//Add properties
		while (it.hasNext()) {
			Map.Entry pair = (Map.Entry) it.next();
			String key = pair.getKey().toString();
			String value = pair.getValue().toString();
			if(value.length()>500){
				Text t=new Text(value);
				e.setProperty(key, t);
			}else{			
				e.setProperty(key, value);
			}
		}
		
		e.setProperty("apiKey", c.getApiKey());
		
		return e;
	}

	@Override
	public List<MoraxusKey> getChildrenKeys(MoraxusKey parentKey) {
		List<MoraxusKey> children=new ArrayList<MoraxusKey>();
		
		Query q = new Query("parentKey_key");
		q.addFilter("parentKey", Query.FilterOperator.EQUAL, parentKey.toString());
		
		PreparedQuery pq = datastore.prepare(q);

		for (Entity result : pq.asIterable()) {
			String keyString = (String) result.getProperty("key");
			children.add(new MoraxusKey(keyString));
		}	
		
		return children;
	}

	@Override
	public MoraxusKey saveObject(Content c) {
		Key key=datastore.put(contentToEntity(c));
		return new MoraxusKey(KeyFactory.keyToString(key));
	}

	@Override
	public boolean hasReadAccess(String userId, String contentId) {
		Query q = new Query("readAccess");
		q.addFilter("userId", Query.FilterOperator.EQUAL, userId);
		q.addFilter("contentId", Query.FilterOperator.EQUAL, contentId);
		
		PreparedQuery pq = datastore.prepare(q);

		for (Entity result : pq.asIterable()) {
			return true;
		}	
		
		return false;
	}
	
	@Override
	public boolean hasWriteAccess(String userId, String contentId) {
		Query q = new Query("writeAccess");
		q.addFilter("userId", Query.FilterOperator.EQUAL, userId);
		q.addFilter("contentId", Query.FilterOperator.EQUAL, contentId);
		
		PreparedQuery pq = datastore.prepare(q);

		for (Entity result : pq.asIterable()) {
			return true;
		}	
		
		return false;
	}

	@Override
	public void setReadAccess(String userId, String contentId) {
		Entity e=new Entity("readAccess");
		e.setProperty("userId", userId);
		e.setProperty("contentId", contentId);
		
		datastore.put(e);
	}

	@Override
	public void setWriteAccess(String userId, String contentId) {
		Entity e=new Entity("writeAccess");
		e.setProperty("userId", userId);
		e.setProperty("contentId", contentId);
		
		datastore.put(e);
	}

	@Override
	public void delete(MoraxusKey key) {
		datastore.delete(KeyFactory.stringToKey(key.toString()));
		//TODO: Delete parent/child relations
	}
}
